"""
API for procedural generation of stereolitography files, a.k.a. STL.

Notes:

- generates plain text STL files that aren't exactly standard, but can
be parsed just fine by modern tools;
- uses a right-handed coordinate system by default: you can interpret
the y axis as going into the screen and the z axis upwards, or else y
as going upwards and z as coming towards you (both can make sense);
- all transformations are made around the origin: you might want to
scale and rotate your shapes before translating them into position.
- all angles are given in radians -- make sure to convert from degrees
if necessary.

Py2STL is deliberately very simple, not unlike a child's toy blocks;
the whole point is not requiring a big, integrated system just to spit
out a few shapes.
"""

from __future__ import division

from math import sin, cos, pi

class Vertex:
	__slots__ = 'x', 'y', 'z'

	def __init__(self, x, y, z):
		self.x = x
		self.y = y
		self.z = z
	
	def translateX(self, x):
		self.x += x
		return self
	
	def translateY(self, y):
		self.y += y
		return self
	
	def translateZ(self, z):
		self.z += z
		return self
		
	def translate(self, x, y, z):
		self.x += x
		self.y += y
		self.z += z
		return self
	
	def scaleX(self, x):
		self.x *= x
		return self
	
	def scaleY(self, y):
		self.y *= y
		return self
	
	def scaleZ(self, z):
		self.z *= z
		return self
		
	def scale(self, x, y, z):
		self.x *= x
		self.y *= y
		self.z *= z
		return self
	
	def rotateX(self, x):
		y, z = self.y, self.z
		self.y = y * cos(x) - z * sin(x)
		self.z = y * sin(x) + z * cos(x)
		return self
	
	def rotateY(self, y):
		x, z = self.x, self.z
		self.x = x * cos(y) - z * sin(y)
		self.z = x * sin(y) + z * cos(y)
		return self
	
	def rotateZ(self, z):
		x, y = self.x, self.y
		self.x = x * cos(z) - y * sin(z)
		self.y = x * sin(z) + y * cos(z)
		return self
		
	def rotate(self, x, y, z):
		self.rotateX(x)
		self.rotateY(y)
		self.rotateZ(z)
		return self
	
	def shearX(self, y, z):
		self.x += self.y * y + self.z * z
		return self
	
	def shearY(self, x, z):
		self.y += self.x * x + self.z * z
		return self
	
	def shearZ(self, x, y):
		self.z += self.x * x + self.y * y
		return self
	
	def distortX(self, y, z):
		self.y *= abs(self.x) * y
		self.z *= abs(self.x) * z
		return self
	
	def distortY(self, x, z):
		self.x *= abs(self.y) * x
		self.z *= abs(self.y) * z
		return self
	
	def distortZ(self, x, y):
		self.x *= abs(self.z) * x
		self.y *= abs(self.z) * y
		return self
	
	def __str__(self):
		return "vertex {0} {1} {2}".format(self.x, self.y, self.z)

class Shape:
	def __init__(self, name=""):
		self.name = name
		self.vertices = []
		self.faces = []
	
	def translateX(self, x):
		for i in self.vertices:
			i.x += x
		return self
	
	def translateY(self, y):
		for i in self.vertices:
			i.y += y
		return self
	
	def translateZ(self, z):
		for i in self.vertices:
			i.z += z
		return self
		
	def translate(self, x, y, z):
		for i in self.vertices:
			i.translate(x, y, z)
		return self
	
	def scaleX(self, x):
		for i in self.vertices:
			i.x *= x
		return self
	
	def scaleY(self, y):
		for i in self.vertices:
			i.y *= y
		return self
	
	def scaleZ(self, z):
		for i in self.vertices:
			i.z *= z
		return self
		
	def scale(self, x, y, z):
		for i in self.vertices:
			i.scale(x, y, z)
		return self
	
	def rotateX(self, x):
		for i in self.vertices:
			i.rotateX(x)
		return self
	
	def rotateY(self, y):
		for i in self.vertices:
			i.rotateY(y)
		return self
	
	def rotateZ(self, z):
		for i in self.vertices:
			i.rotateZ(z)
		return self
		
	def rotate(self, x, y, z):
		for i in self.vertices:
			i.rotate(x, y, z)
		return self
	
	def shearX(self, y, z):
		for i in self.vertices:
			i.shearX(y, z)
		return self
	
	def shearY(self, x, z):
		for i in self.vertices:
			i.shearY(x, z)
		return self
	
	def shearZ(self, x, y):
		for i in self.vertices:
			i.shearZ(x, y)
		return self
	
	def distortX(self, y, z):
		for i in self.vertices:
			i.distortX(y, z)
		return self
	
	def distortY(self, x, z):
		for i in self.vertices:
			i.distortY(x, z)
		return self
	
	def distortZ(self, x, y):
		for i in self.vertices:
			i.distortZ(x, y)
		return self

	def __str__(self):
		text = "solid {0}\n".format(self.name)
		for i in self.faces:
			text += "facet normal 0 0 0\n"
			text += "\touter loop\n"
			for j in i:
				v = self.vertices[j]
				text += "\t\t{0}\n".format(str(v))
			text += "\tendloop\n"
			text += "endfacet\n"
		text += "endsolid {0}".format(self.name)
		return text

class Group:
	def __init__(self):
		self.members = []
	
	def translateX(self, x):
		for i in self.members:
			i.translateX(x)
		return self
	
	def translateY(self, y):
		for i in self.members:
			i.translateY(y)
		return self
	
	def translateZ(self, z):
		for i in self.members:
			i.translateZ(z)
		return self
		
	def translate(self, x, y, z):
		for i in self.members:
			i.translate(x, y, z)
		return self
	
	def scaleX(self, x):
		for i in self.members:
			i.scaleX(x)
		return self
	
	def scaleY(self, y):
		for i in self.members:
			i.scaleY(y)
		return self
	
	def scaleZ(self, z):
		for i in self.members:
			i.scaleZ(z)
		return self
		
	def scale(self, x, y, z):
		for i in self.members:
			i.scale(x, y, z)
		return self
	
	def rotateX(self, x):
		for i in self.members:
			i.rotateX(x)
		return self
	
	def rotateY(self, y):
		for i in self.members:
			i.rotateY(y)
		return self
	
	def rotateZ(self, z):
		for i in self.members:
			i.rotateZ(z)
		return self
		
	def rotate(self, x, y, z):
		for i in self.members:
			i.rotate(x, y, z)
		return self
	
	def shearX(self, y, z):
		for i in self.members:
			i.shearX(y, z)
		return self
	
	def shearY(self, x, z):
		for i in self.members:
			i.shearY(x, z)
		return self
	
	def shearZ(self, x, y):
		for i in self.members:
			i.shearZ(x, y)
		return self
	
	def distortX(self, y, z):
		for i in self.members:
			i.distortX(y, z)
		return self
	
	def distortY(self, x, z):
		for i in self.members:
			i.distortY(x, z)
		return self
	
	def distortZ(self, x, y):
		for i in self.members:
			i.distortZ(x, y)
		return self

	def append(self, member):
		self.members.append(member)
		return self

	def __iadd__(self, member):
		self.members.append(member)
		return self

	def __len__(self):
		return len(self.members)
	
	def __getitem__(self, key):
		return self.members[key]
	
	def __setitem__(self, key, value):
		self.members[key] = value
	
	def __delitem__(self, key):
		del self.members[key]

	def __str__(self):
		return "\n".join([str(m) for m in self.members])

def sheet(width=2, height=2):
	"""Make a simple rectangle in the x-y plane, centered on origin."""
	shape = Shape()
	shape.vertices.append(Vertex(-1, -1, 0))
	shape.vertices.append(Vertex(1, -1, 0))
	shape.vertices.append(Vertex(1, 1, 0))
	shape.vertices.append(Vertex(-1, 1, 0))
	shape.faces.append((0, 1, 2))
	shape.faces.append((0, 2, 3))
	shape.scale(width / 2, height / 2, 1)
	return shape

def disc(radius=1, segments=12):
	"""
	Make a disc of given radius -- basically a many-sided polygon.
	Implemented as a perfectly flat cone.
	"""
	return cone(0, radius, segments)

def cone(h=2, r=1, segments=12):
	"""An uncapped cone of given height and radius, pointed along Z."""
	shape = Shape()
	shape.vertices.append(Vertex(0, 0, h / 2))
	step = 2 * pi / segments
	i = 0
	while i < 2 * pi:
		shape.vertices.append(Vertex(cos(i) * r, sin(i) * r, -h / 2))
		i += step
	for i in range(1, len(shape.vertices) - 1):
		shape.faces.append((0, i, i + 1))
	shape.faces.append((0, i + 1, 1))
	return shape

def cylinder(height=2, radius=1, segments=12):
	"""A Bucket (truncated cone) with the same radius at both ends."""
	return bucket(height, radius, radius, segments)

def bucket(h=2, r1=1, r2=1, segments=12):
	"""A truncated cone of given height and radii, open at both ends."""
	shape = Shape()
	step = 2 * pi / segments
	i = 0
	while i < 2 * pi:
		shape.vertices.append(
			Vertex(cos(i) * r2, sin(i) * r2, -h / 2))
		i += step
	i = 0
	while i < 2 * pi:
		shape.vertices.append(
			Vertex(cos(i) * r1, sin(i) * r1, h / 2))
		i += step
	for i in range(0, segments - 1):
		shape.faces.append((i, i + 1, i + 1 + segments))
		shape.faces.append((i, i + 1 + segments, i + segments))
	shape.faces.append((segments - 1, 0, segments))
	shape.faces.append((segments - 1, segments, segments * 2 - 1))
	return shape

def prism(height=2, radius1=1, radius2=1):
	"""Make a Bucket with just three segments, capped at both ends."""
	shape = bucket(height, radius1, radius2, 3)
	shape.faces.append((0, 2, 1))
	shape.faces.append((3, 4, 5))
	return shape

def simplex(height=2, radius=1):
	"""A tetrahedron (a cone with just three segments and a bottom)."""
	shape = cone(height, radius, 3)
	shape.faces.append((1, 3, 2))
	return shape

def pyramid(x=2, y=2, z=2):
	"""Make a pyramid of given size (with a bottom)."""
	shape = Shape()

	shape.vertices.append(Vertex(0, 0, 1))
	
	shape.vertices.append(Vertex(-1, -1, -1))
	shape.vertices.append(Vertex(1, -1, -1))
	shape.vertices.append(Vertex(1, 1, -1))
	shape.vertices.append(Vertex(-1, 1, -1))
	
	shape.faces.append((0, 1, 2))
	shape.faces.append((0, 2, 3))
	shape.faces.append((0, 3, 4))
	shape.faces.append((0, 4, 1))
	# Bottom side
	shape.faces.append((1, 4, 3))
	shape.faces.append((1, 3, 2))

	shape.scale(x / 2, y / 2, z / 2)
	return shape

def octahedron(x=2, y=2, z=2):
	"""Make an octahedron of given width, depth and height."""
	shape = Shape()

	shape.vertices.append(Vertex(0, 0, 1))
	
	shape.vertices.append(Vertex(-1, -1, 0))
	shape.vertices.append(Vertex(1, -1, 0))
	shape.vertices.append(Vertex(1, 1, 0))
	shape.vertices.append(Vertex(-1, 1, 0))

	shape.vertices.append(Vertex(0, 0, -1))
	
	# Top half
	shape.faces.append((0, 1, 2))
	shape.faces.append((0, 2, 3))
	shape.faces.append((0, 3, 4))
	shape.faces.append((0, 4, 1))
	# Bottom half
	shape.faces.append((5, 2, 1))
	shape.faces.append((5, 3, 2))
	shape.faces.append((5, 4, 3))
	shape.faces.append((5, 1, 4))

	shape.scale(x / 2, y / 2, z / 2)
	return shape

def box(x=2, y=2, z=2):
	"""Make a cuboid of given width, depth and height."""
	shape = Shape()
	
	shape.vertices.append(Vertex(-1, -1, -1))
	shape.vertices.append(Vertex(-1, 1, -1))
	shape.vertices.append(Vertex(1, 1, -1))
	shape.vertices.append(Vertex(1, -1, -1))
	
	shape.vertices.append(Vertex(-1, -1, 1))
	shape.vertices.append(Vertex(1, -1, 1))
	shape.vertices.append(Vertex(1, 1, 1))
	shape.vertices.append(Vertex(-1, 1, 1))
	
	# Bottom side
	shape.faces.append((0, 1, 3))
	shape.faces.append((1, 2, 3))
	# Top side
	shape.faces.append((4, 5, 7))
	shape.faces.append((5, 6, 7))
	# Front side
	shape.faces.append((0, 3, 4))
	shape.faces.append((3, 5, 4))
	# Back side
	shape.faces.append((1, 7, 2))
	shape.faces.append((2, 7, 6))
	# Left side
	shape.faces.append((0, 4, 1))
	shape.faces.append((1, 4, 7))
	# Right side
	shape.faces.append((2, 6, 5))
	shape.faces.append((2, 5, 3))

	shape.scale(x / 2, y / 2, z / 2)
	return shape
