Py2STL
======


Welcome to Py2STL, a Python library for easily generating 3D geometry from a script and saving it in STL format. It's good for making quick-and-dirty visualizations, or perhaps crank out repetitive elements to import into larger scenes. Mostly, though, it's just fun to work on.

How to use
----------

The simplest thing you can possibly do with Py2STL is to run the following command in a terminal:

	python -c 'from py2stl import *; print(sheet())'

which will output a little square:

	solid 
	facet normal 0 0 0
		outer loop
			vertex -1.0 -1.0 0
			vertex 1.0 -1.0 0
			vertex 1.0 1.0 0
		endloop
	endfacet
	facet normal 0 0 0
		outer loop
			vertex -1.0 -1.0 0
			vertex 1.0 1.0 0
			vertex -1.0 1.0 0
		endloop
	endfacet
	endsolid

But to properly take advantage of it you'll have to write a script like this one:

	#!/usr/bin/env python3
	from __future__ import print_function

	from math import radians
	from py2stl import *

	for i in range(20):
		print(box(1, 0.5, 0.3)
			.translate(0.5, 0, i * 0.25)
			.rotateZ(i * radians(18)))

(Don't worry, it should run just as well in Python 2.7; I happen to use 3.3 by default.)

To create the STL model, just run the script and redirect the output to a file. Some (older) viewers will only see the first shape in it; I'll try to add a merging feature if there's interest.

Features
--------

Py2STL has 10 primitives (though several are special cases of others), emitted by one of these functions:

	sheet(width=2, height=2) # A simple rectangle.
	disc(radius=1, segments=12)
	cone(h=2, r=1, segments=12)
	cylinder(height=2, radius=1, segments=12)
	bucket(h=2, r1=1, r2=1, segments=12) # A truncated cone.
	prism(height=2, radius1=1, radius2=1)
	simplex(height=2, radius=1) # A tetrahedron.
	pyramid(x=2, y=2, z=2)
	octahedron(x=2, y=2, z=2)
	box(x=2, y=2, z=2)

(Incompatible change: as of 21 January 2017, all the primitives take optional size arguments; scripts only using the defaults should be unaffected.)

Each primitive is an instance of class `Shape`, and can be transformed with one of the following methods:

	translateX(x)
	translateY(y)
	translateZ(z)
	translate(x, y, z)

	scaleX(x)
	scaleY(y)
	scaleZ(z)
	scale(x, y, z)

	rotateX(x)
	rotateY(y)
	rotateZ(z)
	rotate(x, y, z)
	
	shearX(y, z)
	shearY(x, z)
	shearZ(x, y)
	
	distortX(y, z)
	distortY(x, z)
	distortZ(x, y)

Note that all transformations take place around the origin. You might want to rotate and scale before translating.

To manipulate several shapes at once, you can add them to an instance of class `Group`, either with the `append()` method or simply the `+=` operator. Groups also support all the same transformations, so you can nest them inside each other just fine. See the `plateau.py` example.

Roadmap
-------

- Figure out a good way to cap cones and cylinders.
- Find or figure out a good way to generate faces for a sphere primitive (the vertices are easy enough).

Similar software
----------------

- The [Three.js library](https://threejs.org/) has a similar API, that's very useful for putting placeholder art in a game, but I don't think the results can be easily exported.
- [OpenJSCAD](http://openjscad.org/) is the same concept, on steroids -- a kind of "POV-Ray for polygons" if you like. It's much more capable, but also not nearly as easy to use.
