from __future__ import division

import py2stl

class Camera:
	__slots__ = 'x', 'y', 'z', 'f', 'px', 'py', 'scale'

	def __init__(self, x, y, z, f):
		self.x = x
		self.y = y
		self.z = z
		self.f = f
		
		self.px = 0
		self.py = 0
		self.scale = 0
	
	def project(self, x, y, z):
		x -= self.x
		y -= self.y
		z -= self.z
		if z != 0:
			self.scale = self.f / z
			self.px = x * self.scale
			self.py = y * self.scale
		else:
			self.scale = 0
			self.px = 0
			self.py = 0

class Vertex(py2stl.Vertex):
	def __add__(self, v):
		return Vertex(self.x + v.x, self.y + v.y, self.z + v.z)

	def __sub__(self, v):
		return Vertex(self.x - v.x, self.y - v.y, self.z - v.z)

	def __truediv__(self, n):
		return Vertex(self.x / n, self.y / n, self.z / n)

class Shape(py2stl.Shape):
	@classmethod
	def make_from(cls, other):
		ns = Shape(other.name)
		for i in other.vertices:
			ns.vertices.append(Vertex(i.x, i.y, i.z))
		for i in other.faces:
			ns.faces.append(tuple(i))
		return ns
	
	def copy(self):
		ns = Shape(self.name)
		for i in self.vertices:
			ns.vertices.append(Vertex(i.x, i.y, i.z))
		for i in self.faces:
			ns.faces.append(tuple(i))
		return ns
	
	def render_through(self, cam, renderer):
		proj = []
		for i in self.vertices:
			# Convert right-hand coords to left hand.
			cam.project(i.x, i.z, i.y)
			proj.append((cam.px, cam.py, i.y > cam.z + 0.001))
		for i in self.faces:
			if len(i) == 3:
				v0 = proj[i[0]]
				v1 = proj[i[1]]
				v2 = proj[i[2]]
				# Skip edges with an end behind the camera.
				if v0[2] and v1[2]:
					renderer(v0[0], v0[1], v1[0], v1[1])
				if v1[2] and v2[2]:
					renderer(v1[0], v1[1], v2[0], v2[1])
				if v2[2] and v0[2]:
					renderer(v2[0], v2[1], v0[0], v0[1])
			else: # Assume a quad to avoid testing again.
				v0 = proj[i[0]]
				v1 = proj[i[1]]
				v2 = proj[i[2]]
				v3 = proj[i[3]]
				# Skip edges with an end behind the camera.
				if v0[2] and v1[2]:
					renderer(v0[0], v0[1], v1[0], v1[1])
				if v1[2] and v2[2]:
					renderer(v1[0], v1[1], v2[0], v2[1])
				if v2[2] and v3[2]:
					renderer(v2[0], v2[1], v3[0], v3[1])
				if v3[2] and v0[2]:
					renderer(v3[0], v3[1], v0[0], v0[1])

	def bbox_min(self):
		if len(self.vertices) == 0:
			return Vertex(0, 0, 0)
		else:
			min_x = float("inf")
			min_y = float("inf")
			min_z = float("inf")
			for i in self.vertices:
				if i.x < min_x:
					min_x = i.x
				if i.y < min_y:
					min_y = i.y
				if i.z < min_z:
					min_z = i.z
			return Vertex(min_x, min_y, min_z)

	def bbox_max(self):
		if len(self.vertices) == 0:
			return Vertex(0, 0, 0)
		else:
			max_x = float("-inf")
			max_y = float("-inf")
			max_z = float("-inf")
			for i in self.vertices:
				if i.x > max_x:
					max_x = i.x
				if i.y > max_y:
					max_y = i.y
				if i.z > max_z:
					max_z = i.z
			return Vertex(max_x, max_y, max_z)

	def vertex_count(self):
		return len(self.vertices)

	def face_count(self):
		return len(self.faces)

class Group(py2stl.Group):
	def copy(self):
		ng = Group()
		ng.members = [i.copy() for i in self.members]
		return ng
	
	def render_through(self, camera, renderer):
		for i in self.members:
			i.render_through(camera, renderer)
	
	def bbox_min(self):
		if len(self.members) == 0:
			return Vertex(0, 0, 0)
		else:
			min_x = float("inf")
			min_y = float("inf")
			min_z = float("inf")
			for i in self.members:
				bb_min = i.bbox_min()
				if bb_min.x < min_x:
					min_x = bb_min.x
				if bb_min.y < min_y:
					min_y = bb_min.y
				if bb_min.z < min_z:
					min_z = bb_min.z
			return Vertex(min_x, min_y, min_z)
	
	def bbox_max(self):
		if len(self.members) == 0:
			return Vertex(0, 0, 0)
		else:
			max_x = float("-inf")
			max_y = float("-inf")
			max_z = float("-inf")
			for i in self.members:
				bb_max = i.bbox_max()
				if bb_max.x > max_x:
					max_x = bb_max.x
				if bb_max.y > max_y:
					max_y = bb_max.y
				if bb_max.z > max_z:
					max_z = bb_max.z
			return Vertex(max_x, max_y, max_z)

	def vertex_count(self):
		return sum((i.vertex_count() for i in self.members))

	def face_count(self):
		return sum((i.face_count() for i in self.members))

class STLparser:
	def __init__(self, source):
		self.source = source
		self.advance_line()
	
	def advance_line(self):
		self.line = next(self.source, None)
		if self.line != None:
			self.line = self.line.strip()
	
	def match_start(self, text):
		if self.line == None:
			return False
		elif self.line.startswith(text):
			self.advance_line()
			return True
		else:
			return False
	
	def match_line(self, text):
		if self.line == text:
			self.advance_line()
			return True
		else:
			return False
	
	def parse_solid(self, shape):
		if self.line == None:
			return False
		elif self.match_start("solid"):
			self.vertices = {}
			while not self.match_start("endsolid"):
				self.parse_facet(shape)
			return True
		else:
			raise SyntaxError("STL solid expected")

	def parse_facet(self, shape):
		if self.line == None:
			raise SyntaxError("Unexpected end of input in solid")
		elif not self.match_start("facet"):
			raise SyntaxError("STL facet expected")
		elif not self.match_line("outer loop"):
			raise SyntaxError("STL outer loop expected")
		
		while not self.match_line("endloop"):
			facet = []
			while self.line.startswith("vertex"):
				facet.append(self.parse_vertex(shape))
				self.advance_line()
			shape.faces.append(tuple(facet))
			if self.line != "endloop":
				raise SyntaxError("STL endloop expected")

		if not self.match_line("endfacet"):
			raise SyntaxError("STL endfacet expected")

	def parse_vertex(self, shape):
		if self.line in self.vertices:
			return self.vertices[self.line]
		else:
			fields = self.line.split()
			x = float(fields[1])
			y = float(fields[2])
			z = float(fields[3])
			v = Vertex(x, y, z)
			shape.vertices.append(v)
			idx = len(shape.vertices) - 1
			self.vertices[self.line] = idx
			return idx

def load_obj(source):
	ns = Shape()
	for i in source:
		fields = i.split()
		if len(fields) == 0:
			continue
		elif fields[0] == "v":
			x = float(fields[1])
			y = float(fields[2])
			z = float(fields[3])
			ns.vertices.append(Vertex(x, y, z))
		elif fields[0] == "f":
			verts = fields[1:]
			face = (int(j.split("/")[0]) - 1 for j in verts)
			ns.faces.append(tuple(face))
	return ns
